/*
  ------------------------------------------------------------------------
  file paramdrivers.C of non-standard ADOL-C version 1.7 as of Feb/09/98
  ------------------------------------------------------------------------
  This file contains definitions for the functions prototyped in
  paramadutils.h.  Each function is an ADOL-C C++ utility.
*/
/* DS. 18.9.95 return type changed to int for forward sweeps and drivers  */


#include "paramadutils.h"   /* Prototypes */
#include "dvlparms.h"  /* Developers Parameters */


#define fabs(x) ((x) > 0 ? (x) : -(x))
#define ceil(x) ((int)((x)+1) - (int)((x) == (int)(x)))


int paramforode(short tag,             // tape identifier
            int n,                 // space dimension + parameters
	    int m,		   // space dimension
            double tau,            // scaling defaults to 1.0
            int dol,               // previous degree defaults to zero
            int deg,               // New degree of consistency
            double** y)            // Taylor series
{
  return paramforodec(tag,n,m,tau,dol,deg,y);
}


int paramforode(short tag, int n, int m, double tau, int deg, double** y)
{
  /***   Default for previous degree is zero, do things from scratch ***/
  int zero = 0;
  return  paramforodec(tag, n, m, tau, zero, deg, y);
}

int paramforode(short tag, int n, int m, int dol, int deg, double** y)
{
  /***   Default for scaling is 1.0       *****/
  double tau = 1.0;
  return paramforodec(tag, n, m, tau, dol, deg, y);
}

int paramforode(short tag, int n, int m, int deg, double** y)
{
  /***    Combination of both previous defaults   ******/
  double tau = 1.0;
  return  paramforode(tag, n, m, tau, deg, y);
}

void paramaccode(int n,             // space dimension plus parameters
	    int m, 	       // space dimension
            double tau,        // scaling defaults to 1.0
            int deg,           // highest degree
	    double*** A,       // input tensor of "partial" Jacobians
            double*** B,       // output tensor of "total" Jacobians
	    short** nonzero )  // optional sparsity characterization 
{
  paramaccodec(n,m,tau,deg,A,B,nonzero);
}

void paramaccode(int n,             // space dimension plus parameters
	    int m, 	       // space dimension
            int deg,           // highest degree
	    double*** A,       // input tensor of "partial" Jacobians
            double*** B,       // output tensor of "total" Jacobians
	    short** nonzero )  // optional sparsity characterization 
{
  double tau = 1.0;
  paramaccodec(n,m,tau,deg,A,B,nonzero);
}

