/*
   --------------------------------------------------------------
   paramtest.C  of non-standard ADOL-C version 1.7 as of Feb/09/98
   --------------------------------------------------------------
*/

#include <stream.h>
#include <stdio.h>

#ifdef __GNUG__
#include <std.h>
#else
#include <stdlib.h>
#endif

/*  This program tests the implementation of the param* routines.  
 *  Essentially, their purpose is to compute the derivatives of Taylor
 *  coefficients of paths with respect to parameters.  The routines do 
 *  this without resorting to augmentation of the phase space.  However,
 *  augmentation should lead to the same results, so this routine checks
 *  that it does. 
 */

#include "adouble.h"    // These includes provide the compiler with
#include "adutils.h"    // definitions and utilities for `adoubles'.
#include "paramadutils.h"  // definitions of param* routines.

#define phdim 2
#define tdim 6
#define DIMLOOP(x) for(x=0;x<tdim;x++)
#define DEGLOOP(x) for(x=0;x<=deg;x++)

extern "C" {

double *x, *zz, *pyp;
double **Taylor, ***Jac, ***A;
short **nonzero;

double *xg, *zzg, *pypg;
double **Taylorg, ***Jacg, ***Ag;
short **nonzerog;

void testode_good(unsigned short tag, double* px, double* pxp, double* p2)
{
/* initialize adoubles */
    adouble xv[6];
    adouble xp[2];
    int i;
 
    trace_on(tag);
    for (i=0;i<2;i++)
        xv[i] <<= px[i];
    for (i=2;i<6;i++)
        xv[i] <<= p2[i-2];

    xp[0] = xv[1];
    xp[1] = -( xv[0]*xv[0]*xv[0] + xv[2]*xv[0]*xv[0] + xv[3]*xv[0] + xv[4] ) + xv[1]*(xv[5] - xv[0]*xv[0]);
 
    for (i=0;i<2;i++)
        xp[i] >>= pxp[i];
 
    trace_off();
 
}

void testode_bad(unsigned short tag, double* px, double* pxp)
{
/* initialize adoubles */
    adouble x[tdim];
    adouble xp[tdim];
    int i;

    trace_on(tag);
    for (i=0;i<tdim;i++)
        x[i] <<= px[i];
 
    xp[0] = x[1];
    xp[1] = -( x[0]*x[0]*x[0] + x[2]*x[0]*x[0] + x[3]*x[0] + x[4] ) + x[1]*(x[5] - x[0]*x[0]);
    xp[2] = 0.0; 
    xp[3] = 0.0; 
    xp[4] = 0.0;
    xp[5] = 0.0;

    for (i=0;i<tdim;i++)
        xp[i] >>= pxp[i];
    trace_off();
}

void allocate_good()
{
    int i;
 
    xg = myalloc1(phdim);
    zzg = myalloc1(tdim);
    pypg = myalloc1(phdim);
    Taylorg = myalloc2(tdim,50);
    Jacg = myalloc3(tdim,phdim,50);
    Ag = myalloc3(tdim,phdim,50);
    nonzerog = new short*[tdim];
    for (i=0;i<tdim;i++)  nonzerog[i] = new short[phdim];
}

void allocate_bad()
{
    int i;
 
    x = myalloc1(tdim);
    zz = myalloc1(tdim);
    pyp = myalloc1(tdim);
    Taylor = myalloc2(tdim,50);
    Jac = myalloc3(tdim,tdim,50);
    A = myalloc3(tdim,tdim,50);
    nonzero = new short*[tdim];
    DIMLOOP(i) nonzero[i] = new short[tdim];
}


int main() 
{
    int deg = 7;
    double *p2,delta;
    int i,j,k;

    p2 = myalloc1(tdim-phdim);

   //  Let's start with doing it the bad way 
/*
    delta = 1.0;
    allocate_bad();
    x[0] = 0.0;
    x[1] = 0.64711;
    x[2] = 0.0;
    x[3] = -1.0;
    x[4] = 0.0;
    x[5] = 0.78;
         
    testode_bad(0,x,pyp);

    for (i=0;i<tdim;i++){
	printf ("%lf %lf\n", x[i],pyp[i]);
        }

    DIMLOOP(i) Taylor[i][0] = x[i];

    forode(0,tdim,delta,deg+1,Taylor);
    reverse(0,tdim,tdim,deg,A,nonzero);
    accode(tdim,delta,deg,A,Jac,nonzero);

    for (k=0;k<deg;k++){

        for (i=0;i<tdim;i++){
            for (j=0;j<tdim;j++){
		printf("%lf ", Jac[i][j][k]);
 		}
	    printf ("    ");
	    for (j=0;j<tdim;j++){
		printf("%lf ", A[i][j][k]);
		}
	    printf("\n");
	    }
	printf (" -- ");

	for (j=0;j<tdim;j++)
	    printf ("%lf ", Taylor[j][k]);
	printf("\n");
	}
*/

    // Now, the good way.  With any luck, they're the same! (more or less)
    allocate_good();
    xg[0] = 0.0;
    xg[1] = 0.64711;
    p2[0] = 0.0;
    p2[1] = -1.0;
    p2[2] = 0.0;
    p2[3] = 0.78;

    testode_good(1,xg,pypg,p2);
/*
    for (i=0;i<phdim;i++){
        printf ("%lf \n", xg[i]);
        }
 */
    for(i=0;i<2;i++) Taylorg[i][0] = xg[i];
    for(i=2;i<6;i++) Taylorg[i][0] = p2[i-phdim];
 
    delta = 1.0;
    paramforode(1,tdim,phdim,delta,deg+1,Taylorg);
    reverse(1,phdim,tdim,deg,Ag,nonzerog);
    paramaccode(tdim,phdim,delta,deg,Ag,Jacg,nonzerog);
 
    for (k=0;k<deg;k++){
/*
        for (i=0;i<phdim;i++){
            for (j=0;j<tdim;j++){
                printf("%lf ", Jacg[i][j][k]);
                }
            printf ("    ");
            for (j=0;j<tdim;j++){
                printf("%lf ", Ag[i][j][k]);
                }
            printf("\n");
            }
        printf (" -- ");
*/
        for (j=0;j<tdim;j++)
            printf ("%lf ", Taylorg[j][k]);
        printf("\n");
        }

    printf("\n");
/*
    for (k=0;k<deg;k++){
        for (i=0;i<phdim;i++){
            for (j=0;j<tdim;j++){
                printf("%1.16lf ", Jacg[i][j][k]-Jac[i][j][k]);
                }
            printf ("    ");
            for (j=0;j<tdim;j++){
                printf("%1.16lf ", Ag[i][j][k]-A[i][j][k]);
                }
            printf("\n");
            }
        printf (" -- ");
        for (j=0;j<tdim;j++)
            printf ("%1.16lf ", Taylorg[j][k]-Taylor[j][k]);
        printf("\n");
        }
*/ 

}

}
