extern "C" {
typedef int bool;
#include <matrix.h>
#include <math.h>
}

#include "adouble.h"
#include "adutils.h"
#include "paramadutils.h"

#define dim 2
#define fulldim 4
#define DIMLOOP(x) for(x=0;x<fulldim;x++)
#define DEGLOOP(x) for(x=0;x<=deg;x++)

void fullcanard(double*,double*,double*); 
void fullcanard(unsigned short,double*,double*,double*);

extern "C" {

	static int fallocated = 0;
	static int allocated = 0;
	static double *x,*params,*pyp;
	static double **Taylor,***Jac,***A;
	static short **nonzero;

	extern void mexErrMsgTxt(const char *);

	void fallocate()
	{
		int i;

		if(fallocated) return;
		x = myalloc1(dim);
		x = myalloc1(dim);
		params = myalloc1(fulldim-dim);
		pyp = myalloc1(dim);
		Taylor = myalloc2(fulldim,50);
		Jac = myalloc3(dim,fulldim,50);
		A = myalloc3(dim,fulldim,50);
		nonzero = new short*[dim];
		for (i=0;i<dim;i++) nonzero[i] = new short[fulldim];

		fallocated = 1;
	}

        void allocate()
        {
                int i;

                if(allocated) return;
                x = myalloc1(dim);
/*                params = myalloc1(fulldim-dim);
                pyp = myalloc1(dim);
                Taylor = myalloc2(fulldim,50);
                Jac = myalloc3(dim,fulldim,50);
                A = myalloc3(dim,fulldim,50);
                nonzero = new short*[dim];
                for (i=0;i<dim;i++) nonzero[i] = new short[fulldim];
*/
                allocated = 1;
        }

	void unallocate()
	{
	
		allocated = 0;
/*		free((char*)x);
 */
		free((char*)params);
		free((char*)pyp);
 
		free((char*)*Taylor); free((char*)Taylor);
		free((char*)**Jac);free((char*)*Jac);free((char*)Jac);
		free((char*)**A);free((char*)*A);free((char*)A);
		free((char*)*nonzero);free((char*)nonzero);
	}

	void mexFunction(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
	{
		double *p1,*p2,*p3,*p4,*p5,delta;
		mxArray *cell_ptr;
		int i,j,deg,k,index;
		int subs[2];

/*		allocate();  

		unallocate();
*/
		fallocate();

		if(nrhs < 3){
			mexErrMsgTxt("Not enough arguments: t(x,params,degree)");
			return;
		}

// Store initial point

		p1 = mxGetPr(prhs[0]);
		DIMLOOP(i) Taylor[i][0] = x[i] = p1[i];

// and parameters
		p2 = mxGetPr(prhs[1]);
		for (i=dim;i<fulldim;i++) Taylor[i][0] = params[i-dim] = p2[i-dim];

// and degree

		p3 = mxGetPr(prhs[2]);
		deg = (int) p3[0];

/*
		printf ("%d ", deg);
		for (i=0;i<4;i++) printf ("%lf ", p2[i]);
		for (i=0;i<2;i++) printf ("%lf ", x[i]);
		printf ("\n");
*/

// ADOL-C calculation

		delta = 1.0;
		fullcanard(1,x,pyp,params);
/*
                for (i=0;i<4;i++) printf ("%lf ", p2[i]);
                for (i=0;i<2;i++) printf ("%lf ", x[i]);
		for (i=0;i<2;i++) printf ("%lf ", pyp[i]);
                printf ("\n");
*/
		paramforode(1,fulldim,dim,delta,deg+1,Taylor);
		reverse(1,dim,fulldim,deg,A);
		paramaccode(fulldim,dim,delta,deg,A,Jac);

/*

		for (i=0;i<deg;i++){
		    printf("%lf %lf\n", Taylor[0][i],Taylor[1][i]);
		    }


                for (i=0;i<dim;i++){
                    for (j=0;j<fulldim;j++)
                        printf ("%lf ", Jac[i][j][5]);
                    printf ("\n");
                    }
		printf ("\n");
*/



// Store taylor series into a MATLAB variable

		plhs[0] = mxCreateDoubleMatrix(deg+1,dim,mxREAL);
		p4 = mxGetPr(plhs[0]);
		for(i=0;i<dim;i++) for(j=0;j<=deg;j++) p4[j+i*(deg+1)] = Taylor[i][j];

// Store Jacobian path into a MATLAB Cell Array

		plhs[1] = mxCreateCellMatrix(1,deg+1);

		for(i=0;i<=deg;i++){
			subs[0] = 0; 
			subs[1] = i;
			index = mxCalcSingleSubscript(plhs[1],2,subs);
			cell_ptr = mxCreateDoubleMatrix(dim,fulldim,mxREAL);
			p5 = mxGetPr(cell_ptr);
			for(j=0; j<dim; j++) for(k=0; k<fulldim; k++) p5[j+k*dim] = Jac[j][k][i];
			mxSetCell(plhs[1],index,cell_ptr);
		}
	}

}
