extern "C" {
typedef int bool;
#include <matrix.h>
#include <math.h>
}

#include "adouble.h"
#include "adutils.h"

#define dim 2
#define DIMLOOP(x) for(x=0;x<dim;x++)
#define DEGLOOP(x) for(x=0;x<=deg;x++)

void field(double*,double*,double);
void field(unsigned short, double*,double*,double);

extern "C" {

	static int allocated = 0;
	static double *x,*zz,*pyp;
	static double **Taylor,***Jac,***A;
	static short **nonzero;

	extern void mexErrMsgTxt(const char *);

	void allocate()
	{
		int i;

		if(allocated) return;
		x = myalloc1(dim);
		zz = myalloc1(dim);
		pyp = myalloc1(dim);
		Taylor = myalloc2(dim,50);
		Jac = myalloc3(dim,dim,50);
		A = myalloc3(dim,dim,50);
		nonzero = new short*[dim];
		DIMLOOP(i) nonzero[i] = new short[dim];

		allocated = 1;
	}

	void mexFunction(int nlhs, mxArray *plhs[], int nrhs, mxArray *prhs[])
	{
		double *p1,*p2,delta,c;
		mxArray *cell_ptr;
		int i,j,deg,k,index;
		int subs[2];

		allocate();

		if(nrhs < 2){
			mexErrMsgTxt("Not enough arguments: t(x,degree)");
			return;
		}

// Store initial point

		p1 = mxGetPr(prhs[0]);
		DIMLOOP(i) Taylor[i][0] = x[i] = p1[i];

// and degree

		p1 = mxGetPr(prhs[1]);
		deg = (int) p1[0];

// and parameter c

		p1 = mxGetPr(prhs[2]);
		c  = p1[0];

// ADOL-C calculation

		delta = 1.0;
		field(0,x,pyp,c);
		forode(0,dim,delta,deg+1,Taylor);
		reverse(0,dim,dim,deg,A,nonzero);
		accode(dim,delta,deg,A,Jac,nonzero);

// Store taylor series into a MATLAB variable

		plhs[0] = mxCreateDoubleMatrix(deg+1,dim,mxREAL);
		p1 = mxGetPr(plhs[0]);
		for(i=0;i<dim;i++) for(j=0;j<=deg;j++) p1[j+i*(deg+1)] = Taylor[i][j];

// Store Jacobian path into a MATLAB Cell Array

		plhs[1] = mxCreateCellMatrix(1,deg+1);

		for(i=0;i<=deg;i++){
			subs[0] = 0; 
			subs[1] = i;
			index = mxCalcSingleSubscript(plhs[1],2,subs);
			cell_ptr = mxCreateDoubleMatrix(dim,dim,mxREAL);
			p1 = mxGetPr(cell_ptr);
			for(j=0; j<dim; j++) for(k=0; k<dim; k++) p1[j+k*dim] = Jac[j][k][i];
			mxSetCell(plhs[1],index,cell_ptr);
		}
	}

}
