% Numerical integrate from mesh point to mesh point around cycle
function [nd, jd] = wcycle(ldata)

global degree c dim cmax epmax hmin param x linc

pt = zeros(1,dim);
nd = ldata(1,:);
pdim = size(param,2);
degm = [1:degree]'*ones(1,dim);

for i=1:(size(ldata,1)-1)
   
%Get mesh point

   x = ldata(i,2:dim+1);

%Set current time and time to next mesh point

   tnow = ldata(i,1);
   tstop = ldata(i+1,1)-ldata(i,1);

%Initialize Jacobian of flow

   jpt = [eye(dim),zeros(dim,pdim)];  

      while tstop > 0

%Call ADOL-C
         
         [tay_pt jmat_pt] = fullt(x,param,degree);

%Estimate time step

         ainc = max(max(log(1+abs(tay_pt(2:degree+1,1:dim)))./degm));
         h = exp(linc - ainc);
         
         if h > tstop - 2*hmin
            h=tstop;
         end;

%Time step
         
         [pt, err_pt] = teval(tay_pt,h);
         em = max(abs(err_pt));

%Check accuracy of time step; recompute if error too large
         
         while em > epmax
            h = 0.7*h;
            if h < hmin
               error('Step size smaller than hmin: the minimum permitted step size');
            end;
            [pt, err_pt] = teval(tay_pt,h);
            em = max(abs(err_pt));
         end;

%Compute Jacobian following time step
         
         dm = (h).^[0:degree];
         j1pt = [eye(dim),zeros(dim,size(param,2))];
         
         for j = 1:degree
            j1pt = j1pt + dm(j+1)*jmat_pt{j};
         end;
         jpt = [j1pt(:,1:dim)*jpt(:,1:dim),j1pt(:,1:dim)*jpt(:,dim+1:dim+pdim)+j1pt(:,dim+1:dim+pdim)];

%Update time and phase space point
                     
         tnow = tnow + h;
         tstop = tstop - h;
         x = pt;
         
      end;

%Reached next mesh point: add point and Jacobian to stored structures

   jd{i} = jpt;
   nd = [nd; ldata(i+1,1), pt];

end;
