%Newton iteration

function [nd1,pmm] = wnewt(ldata)

global degree c dim cmax epmax hmin param x linc pmtol aparam nsteps

nint = size(ldata,1) - 1;

%Compute once around cycle and check accuracy

[nd, jd] = wcycle(ldata);
err = nd(2:nint+1,2:dim+1) - ldata(2:nint+1,2:dim+1);
pmm =  max(max(abs(err)))
if pmm < pmtol
	nd1 = nd; 	
	return;
end

%Compute vector field at mesh points

vfield = zeros(nint+1,dim);
for i = 1:nint+1
        [tay, jmat] = fullt(ldata(i,2:dim+1),param,2);
	vfield(i,:) = tay(2,:);
end

%Begin to build Jacobian JA

JA = zeros(nint*(dim+1),nint*(dim+1)+1);

% time shifts of mesh points

JA(dim*nint+2:(dim+1)*nint,dim*nint+1:(dim+1)*nint-1) = eye(nint-1);
JA(dim*nint+1,dim*nint+1:(dim+1)*nint-1) = -ones(1,nint-1);

% i = 1 rows

JA(1:dim,1:dim) = jd{1}(:,1:dim);
JA(1:dim,dim+1:2*dim) = -eye(dim); 
JA(1:dim,dim*nint+1) = vfield(2,:)';
JA(dim*nint+1,1:dim) = vfield(1,:);
JA(1:dim,nint*(dim+1)+1) = jd{1}(:,dim+aparam);

% intermediate rows

for i = 2:nint-1
	JA((dim)*(i-1)+1:(dim)*i,(dim)*(i-1)+1:(dim)*i) = jd{i}(:,1:dim);
	JA((dim)*(i-1)+1:(dim)*i,(dim)*i+1:(dim)*(i+1)) = -eye(dim);
	JA(dim*(i-1)+1:(dim)*i,nint*(dim+1)+1) = jd{i}(:,dim+aparam);
	JA(nint*dim+i,dim*(i-1)+1:dim*i) = vfield(i,:);
	JA(dim*(i-1)+1:dim*i,nint*dim+i-1) = - vfield(i+1,:)';
	JA(dim*(i-1)+1:dim*i,nint*dim+i) = vfield(i+1,:)';
end;

% i = nint rows

JA((dim)*(nint-1)+1:(dim)*nint,(dim)*(nint-1)+1:(dim)*nint) = jd{nint}(:,1:dim);
JA((dim)*(nint-1)+1:(dim)*nint,1:dim) = -eye(dim);
JA((dim)*(nint-1)+1:(dim)*nint,nint*(dim+1)+1) = jd{nint}(:,dim+aparam);
JA(nint*(dim+1),dim*(nint-1)+1:dim*nint) = vfield(nint,:);
JA(dim*(nint-1)+1:dim*nint,nint*(dim+1)-1) = - vfield(nint+1,:)';
JA(dim*(nint-1)+1:dim*nint,nint*(dim+1)) = vfield(nint+1,:)';

%compute null space for Jacobian

cv = null(JA);

%Form augmented Jacobian: add final row perpendicular to null space

JA3 = [JA;cv'];

%Compute Newton step

errv = [reshape(err',dim*nint,1);zeros(nint+1,1)];
wn_inc = JA3\errv;
pt_inc =  (reshape(wn_inc(1:dim*nint,1),dim,nint))';
t_inc = wn_inc(dim*nint+1:(dim+1)*nint,1);

%Now iterate Newton calculation

for i = 1:nsteps
	
%Update with Newton step

	ldata(1:nint,2:dim+1) = ldata(1:nint,2:dim+1) - pt_inc;
	ldata(nint+1,2:dim+1) = ldata(1,2:dim+1);
	ldata(2:nint+1,1) = ldata(2:nint+1,1) - t_inc;
	param(aparam) = param(aparam)-wn_inc((dim+1)*nint+1);

% Compute next cycle

	[nd, jd] = wcycle(ldata);

%Compute error and return if converged

	err = nd(2:nint+1,2:dim+1) - ldata(2:nint+1,2:dim+1);
	pmm =  max(max(abs(err)))
	if pmm < pmtol
		nd1 = nd; 	
		return;
	end

%Not converged: update Jacobian
	
	for i = 1:nint+1
        	[tay, jmat] = fullt(ldata(i,2:dim+1),param,2);
		vfield(i,:) = tay(2,:);
	end
	for i = 1:nint
		JA3((dim)*(i-1)+1:(dim)*i,(dim)*(i-1)+1:(dim)*i) = jd{i}(:,1:dim);
		JA3((dim)*(i-1)+1:(dim)*i,nint*(dim+1)+1) = jd{i}(:,dim+aparam);
		if i > 1
			JA3(dim*(i-1)+1:dim*i,nint*dim+i-1) = - vfield(i+1,:)';
		end
		JA3(dim*(i-1)+1:dim*i,nint*dim+i) = vfield(i+1,:)';
	end;

%Compute next Newton increment
 
	errv = [reshape(err',dim*nint,1);zeros(nint+1,1)];
	wn_inc = JA3\errv;
	pt_inc =  (reshape(wn_inc(1:dim*nint,1),dim,nint))';
	t_inc = wn_inc(dim*nint+1:(dim+1)*nint,1);
end

