%Refine a mesh, placing new mesh points where Jacobian norm > cmax
function [nd] = wrefine(ldata)

global degree c dim cmax epmax hmin param

pt = zeros(1,dim);
linc = -16*log(10)/degree;
nd = ldata(1,:);
emm = 0;
pdim = size(param,2);
degm = [1:degree]'*ones(1,dim);

for i=1:(size(ldata,1)-1)
   
   x = ldata(i,2:dim+1);
   tnow = ldata(i,1);
   tstop = ldata(i+1,1)-ldata(i,1);
   jpt = [eye(dim),zeros(dim,pdim)];
   
   while tstop > 0
      
      [tay_pt jmat_pt] = fullt(x,param,degree);
      ainc = max(max(log(1+abs(tay_pt(2:degree+1,1:dim)))./degm));
      h = exp(linc - ainc);
      
      if h > tstop - 2*hmin;
         h = tstop;
      end;
      
      [pt, err_pt] = teval(tay_pt,h);
      em = max(abs(err_pt));
      
      while em > epmax
         h = 0.7*h;
         if h < hmin;
            error('Step size smaller than hmin: the minimum permitted step size');
         end;
         [pt, err_pt] = teval(tay_pt,h);
         em = max(abs(err_pt));
      end;
      
      emm = max(em, emm);
      dm = (h).^[0:degree];
      j1pt = [eye(dim),zeros(dim,size(param,2))];
      
      for j = 1:degree
         j1pt = j1pt + dm(j+1)*jmat_pt{j};
      end;
      j2pt = [j1pt(:,1:dim)*jpt(:,1:dim),j1pt(:,1:dim)*jpt(:,dim+1:dim+pdim)+j1pt(:,dim+1:dim+pdim)];
      
      if tnow > nd(size(nd,1),1) + hmin
         if (norm(j2pt(1:dim,1:dim)) > cmax)  
            nd = [nd; tnow, x];
            %disp('Jacobian too large: add mesh point to nd');  
         end;
      end;
      
      tnow = tnow + h;
      jpt = j2pt;
      tstop = tstop - h;
      x = pt;
      
   end;
   nd = [nd; tnow, pt];
end;
